from django.core.exceptions import PermissionDenied
from .apps import CALCULATION_RULES
from .calculation_rule import ContributionValuationRule


def get_rule_name(class_name):
    list_rule_name = []
    for calculation_rule in CALCULATION_RULES:
        result_signal = calculation_rule.signal_get_rule_name.send(sender=class_name, class_name=class_name)
        if result_signal:
            list_rule_name.extend(result_signal)
    return list_rule_name


def get_rule_details(class_name):
    list_rule_details = []
    for calculation_rule in CALCULATION_RULES:
        result_signal = calculation_rule.signal_get_rule_details.send(sender=class_name, class_name=class_name)
        if result_signal:
            list_rule_details.extend(result_signal)
    return list_rule_details

def get_calculation_object(uuid):
    for calculation_rule in CALCULATION_RULES:
        if str(calculation_rule.uuid) == str(uuid):
            return calculation_rule

def run_calculation_rules(instance, context, user, **kwargs):
    for calculation_rule in CALCULATION_RULES:
        result_signal = calculation_rule.signal_calculate_event.send(
            sender=instance.__class__.__name__, instance=instance, user=user, context=context, **kwargs
        )
        if result_signal[0][1]:
            return result_signal
    # if no listened calculation rules - return None
    return None


def get_parameters(class_name, instance):
    """ className is the class name of the object where the calculation param need to be added
        instance is where the link with a calculation need to be found,
         like the CPB in case of PH insuree or Contract Details
    """
    list_parameters = []
    for calculation_rule in CALCULATION_RULES:
        result_signal = calculation_rule.signal_get_param.send(
            sender=instance, class_name=class_name, instance=instance
        )
        if result_signal:
            list_parameters.extend(result_signal)
    # return the ruleDetails that are valid to classname and related to instance
    return list_parameters


def get_linked_class(class_name_list=None):
    return_list_class = []
    for calculation_rule in CALCULATION_RULES:
        if class_name_list == None:
            result_signal = calculation_rule.signal_get_linked_class.send(sender="None", class_name=None)
            if result_signal:
                return_list_class.extend(result_signal)
        else:
            for class_name in class_name_list:
                result_signal = calculation_rule.signal_get_linked_class.send(sender=class_name, class_name=class_name)
                if result_signal:
                    return_list_class.extend(result_signal)
    return return_list_class



import logging

from django.db import migrations

logger = logging.getLogger(__name__)


MIGRATION_SQL = """
    /* Calculation */
    DECLARE @SystemRole INT
    SELECT @SystemRole = role.RoleID from tblRole role where IsSystem=64;
    /* Calculation search*/
    IF NOT EXISTS (SELECT * FROM [tblRoleRight] WHERE [RoleID] = @SystemRole AND [RightID] = 153001)
    BEGIN
        INSERT [dbo].[tblRoleRight] ([RoleID], [RightID], [ValidityFrom]) 
        VALUES (@SystemRole, 153001, CURRENT_TIMESTAMP)
    END 
    /* Calculation update*/
    IF NOT EXISTS (SELECT * FROM [tblRoleRight] WHERE [RoleID] = @SystemRole AND [RightID] = 153003)
    BEGIN
        INSERT [dbo].[tblRoleRight] ([RoleID], [RightID], [ValidityFrom]) 
        VALUES (@SystemRole, 153003, CURRENT_TIMESTAMP)
    END 
"""


class Migration(migrations.Migration):
    dependencies = [
        ('calculation', '0002_auto_20210118_1426')
    ]

    operations = [
        migrations.RunSQL(sql=MIGRATION_SQL),
    ]



import json

from .apps import AbsCalculationRule
from .config import CLASS_RULE_PARAM_VALIDATION, CLASS_RULE_PARAM_VALIDATION2, \
    DESCRIPTION_CONTRIBUTION_VALUATION, FROM_TO
from contribution_plan.models import ContributionPlanBundleDetails
from core.signals import Signal
from core import datetime
from django.contrib.contenttypes.models import ContentType
from policyholder.models import PolicyHolderInsuree


class ContributionValuationRule(AbsCalculationRule):
    version = 1
    uuid = "0e1b6dd4-04a0-4ee6-ac47-2a99cfa5e9a8"
    calculation_rule_name = "CV: percent of income"
    description = DESCRIPTION_CONTRIBUTION_VALUATION
    impacted_class_parameter = CLASS_RULE_PARAM_VALIDATION
    date_valid_from = datetime.datetime(2000, 1, 1)
    date_valid_to = None
    status = "active"
    from_to = FROM_TO
    type = "account_receivable"
    sub_type = "contribution"

    signal_get_rule_name = Signal(providing_args=[])
    signal_get_rule_details = Signal(providing_args=[])
    signal_get_param = Signal(providing_args=[])
    signal_get_linked_class = Signal(providing_args=[])
    signal_calculate_event = Signal(providing_args=[])
    signal_convert_from_to = Signal(providing_args=[])

    @classmethod
    def ready(cls):
        now = datetime.datetime.now()
        condition_is_valid = (cls.date_valid_from <= now <= cls.date_valid_to) \
            if cls.date_valid_to else (now >= cls.date_valid_from and cls.date_valid_to is None)
        if condition_is_valid:
            if cls.status == "active":
                # register signals getParameter to getParameter signal and getLinkedClass ot getLinkedClass signal
                cls.signal_get_rule_name.connect(cls.get_rule_name, dispatch_uid="on_get_rule_name_signal")
                cls.signal_get_rule_details.connect(cls.get_rule_details, dispatch_uid="on_get_rule_details_signal")
                cls.signal_get_param.connect(cls.get_parameters, dispatch_uid="on_get_param_signal")
                cls.signal_get_linked_class.connect(cls.get_linked_class, dispatch_uid="on_get_linked_class_signal")
                cls.signal_calculate_event.connect(cls.run_calculation_rules, dispatch_uid="on_calculate_event_signal")

    @classmethod
    def active_for_object(cls, instance, context, type='account_receivable', sub_type='contribution'):
        return instance.__class__.__name__ == "ContractContributionPlanDetails" \
               and context in ["create", "update"] \
               and cls.check_calculation(instance)

    @classmethod
    def check_calculation(cls, instance):
        match = False
        class_name = instance.__class__.__name__
        list_class_name = [
            "PolicyHolder", "ContributionPlan",
            "PolicyHolderInsuree", "ContractDetails",
            "ContractContributionPlanDetails", "ContributionPlanBundle"
        ]
        if class_name == "ABCMeta":
            match = str(cls.uuid) == str(instance.uuid)
        elif class_name == "ContributionPlan":
            match = str(cls.uuid) == str(instance.calculation)
        elif class_name == "ContributionPlanBundle":
            list_cpbd = list(ContributionPlanBundleDetails.objects.filter(
                contribution_plan_bundle=instance
            ))
            for cpbd in list_cpbd:
                if match is False:
                    if cls.check_calculation(cpbd.contribution_plan):
                        match = True
        else:
            related_fields = [
                f.name for f in instance.__class__._meta.fields
                if f.get_internal_type() == 'ForeignKey' and f.remote_field.model.__name__ in list_class_name
            ]
            for rf in related_fields:
                match = cls.check_calculation(getattr(instance, rf))
        return match

    @classmethod
    def calculate(cls, instance, **kwargs):
        if instance.__class__.__name__ == "ContractContributionPlanDetails":
            # check type of json_ext - in case of string - json.loads
            cp_params, cd_params = instance.contribution_plan.json_ext, instance.contract_details.json_ext
            ph_insuree = PolicyHolderInsuree.objects.filter(
                insuree=instance.contract_details.insuree).first()
            phi_params = ph_insuree.json_ext
            if isinstance(cp_params, str):
                cp_params = json.loads(cp_params)
            if isinstance(cd_params, str):
                cd_params = json.loads(cd_params)
            if isinstance(phi_params, str):
                phi_params = json.loads(phi_params)
            # check if json external calculation rule in instance exists
            if cp_params:
                cp_params = cp_params["calculation_rule"] if "calculation_rule" in cp_params else None
            if cd_params:
                cd_params = cd_params["calculation_rule"] if "calculation_rule" in cd_params else None
            if phi_params:
                phi_params = phi_params["calculation_rule"] if "calculation_rule" in phi_params else None
            if "rate" in cp_params:
                rate = int(cp_params["rate"])
                if cd_params:
                    if "income" in cd_params:
                        income = float(cd_params["income"])
                    elif "income" in phi_params:
                        income = float(phi_params["income"])
                    else:
                        return False
                elif "income" in phi_params:
                    income = float(phi_params["income"])
                else:
                    return False
                value = float(income) * (rate / 100)
                return value
            else:
                return False
        else:
            return False

    @classmethod
    def get_linked_class(cls, sender, class_name, **kwargs):
        list_class = []
        if class_name is not None:
            model_class = ContentType.objects.filter(model=class_name).first()
            if model_class:
                model_class = model_class.model_class()
                list_class = list_class + \
                             [f.remote_field.model.__name__ for f in model_class._meta.fields
                              if f.get_internal_type() == 'ForeignKey' and f.remote_field.model.__name__ != "User"]
        else:
            list_class.append("Calculation")
        # because we have calculation in ContributionPlan
        #  as uuid - we have to consider this case
        if class_name == "ContributionPlan":
            list_class.append("Calculation")
        # because we have no direct relation in ContributionPlanBundle
        #  to ContributionPlan we have to consider this case
        if class_name == "ContributionPlanBundle":
            list_class.append("ContributionPlan")
        return list_class


class ContributionValuationRule_2(AbsCalculationRule):
    version = 1
    uuid = "0e556dd4-04a0-4aa4-ac47-2a99cfa5e9a8"
    calculation_rule_name = "CV: percent of income, no dependants"
    description = DESCRIPTION_CONTRIBUTION_VALUATION
    impacted_class_parameter = CLASS_RULE_PARAM_VALIDATION2
    date_valid_from = datetime.datetime(2000, 1, 1)
    date_valid_to = None
    status = "active"
    from_to = FROM_TO
    type = "account_receivable"
    sub_type = "contribution"

    signal_get_rule_name = Signal(providing_args=[])
    signal_get_rule_details = Signal(providing_args=[])
    signal_get_param = Signal(providing_args=[])
    signal_get_linked_class = Signal(providing_args=[])
    signal_calculate_event = Signal(providing_args=[])
    signal_convert_from_to = Signal(providing_args=[])

    @classmethod
    def ready(cls):
        now = datetime.datetime.now()
        condition_is_valid = (cls.date_valid_from <= now <= cls.date_valid_to) \
            if cls.date_valid_to else (now >= cls.date_valid_from and cls.date_valid_to is None)
        if condition_is_valid:
            if cls.status == "active":
                # register signals getParameter to getParameter signal and getLinkedClass ot getLinkedClass signal
                cls.signal_get_rule_name.connect(cls.get_rule_name, dispatch_uid="on_get_rule_name_signal")
                cls.signal_get_rule_details.connect(cls.get_rule_details, dispatch_uid="on_get_rule_details_signal")
                cls.signal_get_param.connect(cls.get_parameters, dispatch_uid="on_get_param_signal")
                cls.signal_get_linked_class.connect(cls.get_linked_class, dispatch_uid="on_get_linked_class_signal")
                cls.signal_calculate_event.connect(cls.run_calculation_rules, dispatch_uid="on_calculate_event_signal")

    @classmethod
    def active_for_object(cls, instance, context, type='account_receivable', sub_type='contribution'):
        return instance.__class__.__name__ == "ContractContributionPlanDetails" \
               and context in ["create", "update"] \
               and cls.check_calculation(instance)

    @classmethod
    def check_calculation(cls, instance):
        match = False
        class_name = instance.__class__.__name__
        list_class_name = [
            "PolicyHolder", "ContributionPlan",
            "PolicyHolderInsuree", "ContractDetails",
            "ContractContributionPlanDetails", "ContributionPlanBundle"
        ]
        if class_name == "ABCMeta":
            match = str(cls.uuid) == str(instance.uuid)
        elif class_name == "ContributionPlan":
            match = str(cls.uuid) == str(instance.calculation)
        elif class_name == "ContributionPlanBundle":
            list_cpbd = list(ContributionPlanBundleDetails.objects.filter(
                contribution_plan_bundle=instance
            ))
            for cpbd in list_cpbd:
                if match is False:
                    if cls.check_calculation(cpbd.contribution_plan):
                        match = True
        else:
            related_fields = [
                f.name for f in instance.__class__._meta.fields
                if f.get_internal_type() == 'ForeignKey' and f.remote_field.model.__name__ in list_class_name
            ]
            for rf in related_fields:
                match = cls.check_calculation(getattr(instance, rf))
        return match

    @classmethod
    def calculate(cls, instance, **kwargs):
        if instance.__class__.__name__ == "ContractContributionPlanDetails":
            # check type of json_ext - in case of string - json.loads
            cp_params, cd_params = instance.contribution_plan.json_ext, instance.contract_details.json_ext
            ph_insuree = PolicyHolderInsuree.objects.filter(
                insuree=instance.contract_details.insuree).first()
            phi_params = ph_insuree.json_ext
            if isinstance(cp_params, str):
                cp_params = json.loads(cp_params)
            if isinstance(cd_params, str):
                cd_params = json.loads(cd_params)
            if isinstance(phi_params, str):
                phi_params = json.loads(phi_params)
            # check if json external calculation rule in instance exists
            if cp_params:
                cp_params = cp_params["calculation_rule"] if "calculation_rule" in cp_params else None
            if cd_params:
                cd_params = cd_params["calculation_rule"] if "calculation_rule" in cd_params else None
            if phi_params:
                phi_params = phi_params["calculation_rule"] if "calculation_rule" in phi_params else None
            if "rate" in cp_params:
                rate = int(cp_params["rate"])
                if cd_params:
                    if "income" in cd_params:
                        income = float(cd_params["income"])
                    elif "income" in phi_params:
                        income = float(phi_params["income"])
                    else:
                        return False
                elif "income" in phi_params:
                    income = float(phi_params["income"])
                else:
                    return False
                value = float(income) * (rate / 100)
                return value
            else:
                return False
        else:
            return False

    @classmethod
    def get_linked_class(cls, sender, class_name, **kwargs):
        list_class = []
        if class_name != None:
            model_class = ContentType.objects.filter(model=class_name).first()
            if model_class:
                model_class = model_class.model_class()
                list_class = list_class + \
                             [f.remote_field.model.__name__ for f in model_class._meta.fields
                              if f.get_internal_type() == 'ForeignKey' and f.remote_field.model.__name__ != "User"]
        else:
            list_class.append("Calculation")
        # because we have calculation in ContributionPlan
        #  as uuid - we have to consider this case
        if class_name == "ContributionPlan":
            list_class.append("Calculation")
        # because we have no direct relation in ContributionPlanBundle
        #  to ContributionPlan we have to consider this case
        if class_name == "ContributionPlanBundle":
            list_class.append("ContributionPlan")
        return list_class



# Generated by Django 3.0.3 on 2020-12-04 11:50

import datetime
import dirtyfields.dirtyfields
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import jsonfallback.fields
import simple_history.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='CalculationRules',
            fields=[
                ('id', models.UUIDField(db_column='UUID', default=None, editable=False, primary_key=True, serialize=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', jsonfallback.fields.FallbackJSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', models.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', models.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('date_valid_from', models.DateTimeField(db_column='DateValidFrom', default=datetime.datetime.now)),
                ('date_valid_to', models.DateTimeField(blank=True, db_column='DateValidTo', null=True)),
                ('replacement_uuid', models.UUIDField(db_column='ReplacementUUID', null=True)),
                ('calculation_class_name', models.CharField(blank=True, db_column='CalculationsClassName', max_length=255, null=True)),
                ('description', models.CharField(blank=True, db_column='Description', max_length=255, null=True)),
                ('priority', models.IntegerField(blank=True, db_column='Priority', null=True)),
                ('status', models.IntegerField(blank=True, db_column='Status', null=True)),
                ('user_created', models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='calculationrules_user_created', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='calculationrules_user_updated', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'tblCalculationRules',
            },
            bases=(dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalCalculationRulesDetails',
            fields=[
                ('id', models.UUIDField(db_column='UUID', db_index=True, default=None, editable=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', jsonfallback.fields.FallbackJSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', models.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', models.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('status', models.IntegerField(blank=True, db_column='Status', null=True)),
                ('main', models.BooleanField(blank=True, db_column='Main', null=True)),
                ('params', jsonfallback.fields.FallbackJSONField(blank=True, db_column='Params', null=True)),
                ('class_params', jsonfallback.fields.FallbackJSONField(blank=True, db_column='ClassParams', null=True)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('calculation_rules', models.ForeignKey(blank=True, db_column='CalculationRulesUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='calculation.CalculationRules')),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_created', models.ForeignKey(blank=True, db_column='UserCreatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(blank=True, db_column='UserUpdatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical calculation rules details',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='HistoricalCalculationRules',
            fields=[
                ('id', models.UUIDField(db_column='UUID', db_index=True, default=None, editable=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', jsonfallback.fields.FallbackJSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', models.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', models.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('date_valid_from', models.DateTimeField(db_column='DateValidFrom', default=datetime.datetime.now)),
                ('date_valid_to', models.DateTimeField(blank=True, db_column='DateValidTo', null=True)),
                ('replacement_uuid', models.UUIDField(db_column='ReplacementUUID', null=True)),
                ('calculation_class_name', models.CharField(blank=True, db_column='CalculationsClassName', max_length=255, null=True)),
                ('description', models.CharField(blank=True, db_column='Description', max_length=255, null=True)),
                ('priority', models.IntegerField(blank=True, db_column='Priority', null=True)),
                ('status', models.IntegerField(blank=True, db_column='Status', null=True)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_created', models.ForeignKey(blank=True, db_column='UserCreatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(blank=True, db_column='UserUpdatedUUID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical calculation rules',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='CalculationRulesDetails',
            fields=[
                ('id', models.UUIDField(db_column='UUID', default=None, editable=False, primary_key=True, serialize=False)),
                ('is_deleted', models.BooleanField(db_column='isDeleted', default=False)),
                ('json_ext', jsonfallback.fields.FallbackJSONField(blank=True, db_column='Json_ext', null=True)),
                ('date_created', models.DateTimeField(db_column='DateCreated', null=True)),
                ('date_updated', models.DateTimeField(db_column='DateUpdated', null=True)),
                ('version', models.IntegerField(default=1)),
                ('status', models.IntegerField(blank=True, db_column='Status', null=True)),
                ('main', models.BooleanField(blank=True, db_column='Main', null=True)),
                ('params', jsonfallback.fields.FallbackJSONField(blank=True, db_column='Params', null=True)),
                ('class_params', jsonfallback.fields.FallbackJSONField(blank=True, db_column='ClassParams', null=True)),
                ('calculation_rules', models.ForeignKey(db_column='CalculationRulesUUID', on_delete=django.db.models.deletion.DO_NOTHING, to='calculation.CalculationRules')),
                ('user_created', models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='calculationrulesdetails_user_created', to=settings.AUTH_USER_MODEL)),
                ('user_updated', models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='calculationrulesdetails_user_updated', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'tblCalculationRulesDetails',
            },
            bases=(dirtyfields.dirtyfields.DirtyFieldsMixin, models.Model),
        ),
    ]



import graphene
from .apps import CALCULATION_RULES, CalculationConfig
from .services import get_rule_name, get_parameters, get_linked_class
from django.contrib.contenttypes.models import ContentType


class CalculationRulesGQLType(graphene.ObjectType):
    calculation_class_name = graphene.String()
    status = graphene.String()
    description = graphene.String()
    uuid = graphene.UUID()
    class_param = graphene.JSONString()
    date_valid_from = graphene.Date()
    date_valid_to = graphene.Date()
    from_to = graphene.JSONString()
    type = graphene.String()
    sub_type = graphene.String()


class CalculationRulesListGQLType(graphene.ObjectType):
    calculation_rules = graphene.List(CalculationRulesGQLType)


class LabelParamGQLType(graphene.ObjectType):
    en = graphene.String()
    fr = graphene.String()


class RightParamGQLType(graphene.ObjectType):
    read = graphene.String()
    write = graphene.String()
    update = graphene.String()
    replace = graphene.String()


class OptionParamGQLType(graphene.ObjectType):
    value = graphene.String()
    label = graphene.Field(LabelParamGQLType)


class CalculationParamsGQLType(graphene.ObjectType):
    type = graphene.String()
    name = graphene.String()
    label = graphene.Field(LabelParamGQLType)
    rights = graphene.Field(RightParamGQLType)
    option_set = graphene.List(OptionParamGQLType)
    relevance = graphene.String()
    condition = graphene.String()
    default_value = graphene.String()


class CalculationParamsListGQLType(graphene.ObjectType):
    calculation_params = graphene.List(CalculationParamsGQLType)


class LinkedClassListGQLType(graphene.ObjectType):
    linked_classes = graphene.List(graphene.String)


class Query(graphene.ObjectType):

    calculation_rules_by_class_name = graphene.Field(
        CalculationRulesListGQLType,
        class_name=graphene.Argument(graphene.String, required=True),
    )

    calculation_rules = graphene.Field(
        CalculationRulesListGQLType,
        calculation=graphene.Argument(graphene.UUID, required=False),
        calcrule_type=graphene.Argument(graphene.String, required=False),
    )

    calculation_params = graphene.Field(
        CalculationParamsListGQLType,
        class_name=graphene.Argument(graphene.String, required=True),
        instance_id=graphene.Argument(graphene.String, required=True),
        instance_class_name=graphene.Argument(graphene.String, required=True),
    )

    linked_class = graphene.Field(
        LinkedClassListGQLType,
        class_name_list=graphene.Argument(graphene.List(graphene.String), required=False),
    )

    def resolve_calculation_rules_by_class_name(parent, info, **kwargs):
        if not info.context.user.has_perms(CalculationConfig.gql_query_calculation_rule_perms):
           raise PermissionError("Unauthorized")

        class_name = kwargs.get("class_name", None)
        list_cr = []
        if class_name:
            list_signal_result = get_rule_name(class_name=class_name)
            if list_signal_result:
                for sr in list_signal_result:
                    # get the signal result - calculation rule object
                    #  related to the input class name
                    rule = sr[1]
                    if rule:
                        list_cr.append(
                            CalculationRulesGQLType(
                                calculation_class_name=rule.calculation_rule_name,
                                status=rule.status,
                                description=rule.description,
                                uuid=rule.uuid,
                                class_param=rule.impacted_class_parameter,
                                date_valid_from=rule.date_valid_from,
                                date_valid_to=rule.date_valid_to,
                                from_to=rule.from_to,
                                type=rule.type,
                                sub_type=rule.sub_type
                            )
                        )
        return CalculationRulesListGQLType(list_cr)

    def resolve_calculation_rules(parent, info, **kwargs):
        if not info.context.user.has_perms(CalculationConfig.gql_query_calculation_rule_perms):
            raise PermissionError("Unauthorized")

        calculation = kwargs.get("calculation", None)
        calcrule_type = kwargs.get("calcrule_type", None)

        if calculation or calcrule_type:
            list_cr = []
            for cr in CALCULATION_RULES:
                calculation = f'{calculation}'
                if (cr.uuid == calculation and calcrule_type == cr.type) \
                        or (cr.uuid == calculation and calcrule_type is None) \
                        or (calculation == 'None' and calcrule_type == cr.type):
                    list_cr = _append_to_calcrule_list(list_cr, cr)
        else:
            list_cr = []
            for cr in CALCULATION_RULES:
                list_cr = _append_to_calcrule_list(list_cr, cr)

        return CalculationRulesListGQLType(list_cr)

    def resolve_calculation_params(parent, info, **kwargs):
        if not info.context.user.has_perms(CalculationConfig.gql_query_calculation_rule_perms):
           raise PermissionError("Unauthorized")

        # get the obligatory params from query
        class_name = kwargs.get("class_name", None)
        instance_id = kwargs.get("instance_id", None)
        instance_class_name = kwargs.get("instance_class_name", None)

        instance = None
        # if calculation class - it is not a model
        if instance_class_name == "Calculation":
            for calculation_rule in CALCULATION_RULES:
                if calculation_rule.uuid == instance_id:
                    instance = calculation_rule
        else:
            # get the instance class name to get instance object by uuid
            instance_type = ContentType.objects.get(model=f'{instance_class_name}')
            instance_class = instance_type.model_class()
            instance = instance_class.objects.get(id=instance_id)

        list_params = []
        if class_name and instance:
            # use service to send signal to all class to obtain params related to the instance
            list_signal_result = get_parameters(class_name=class_name, instance=instance)
            tranformed_signal_results = []
            if list_signal_result:
                for sr in list_signal_result:
                    # get the signal result - calculation param object
                    #  related to the input class name and instance
                    # do not include None results
                    if sr[1]:
                        tranformed_signal_results.extend(sr[1])
            # make parameter list from signal unique - discinct list of dict by 'name' keyword
            tranformed_signal_results = list({v['name']: v for v in tranformed_signal_results}.values())
            for param in tranformed_signal_results:
                rights = RightParamGQLType(
                    read=param['rights']['read'] if 'read' in param['rights'] else None,
                    write=param['rights']['write'] if 'write' in param['rights'] else None,
                    update=param['rights']['update'] if 'update' in param['rights'] else None,
                    replace=param['rights']['replace'] if 'replace' in param['rights'] else None,
                )
                #FIXME, either rely on locals (BEST case) or manage it generically
                label = LabelParamGQLType(
                    en=param['label']['en'] if 'en' in param['label'] else None,
                    fr=param['label']['fr'] if 'fr' in param['label'] else None,
                )
                option_set = [OptionParamGQLType(
                    value=ov["value"],
                    label=LabelParamGQLType(en=ov["label"]["en"], fr=ov["label"]["fr"])
                ) for ov in param["optionSet"]] if "optionSet" in param else []

                if "condition" in param:
                    condition = param["condition"] if param["condition"] else None
                else:
                    condition = None
                if "relevance" in param:
                    relevance = param["relevance"] if param["relevance"] else None
                else:
                    relevance = None
 
                list_params.append(
                    CalculationParamsGQLType(
                        type=param['type'],
                        name=param['name'],
                        label=label,
                        rights=rights,
                        option_set=option_set,
                        relevance=relevance,
                        condition=condition,
                        default_value=param['default'] if 'default' in param else "null",
                    )
                )
        return CalculationParamsListGQLType(list_params)

    def resolve_linked_class(parent, info, **kwargs):
        if not info.context.user.has_perms(CalculationConfig.gql_query_calculation_rule_perms):
           raise PermissionError("Unauthorized")
        result_linked_class = []
        # get the params from query
        class_name_list = kwargs.get("class_name_list", None)
        list_signal_result = get_linked_class(class_name_list=class_name_list)
        for sr in list_signal_result:
            if sr[1]:
                result_linked_class = result_linked_class + sr[1]
        result_linked_class = list(set(result_linked_class))
        # remove product when we have PaymentPlan/ContributionPlan object
        # TODO: find a more generic way to avoid loop cause by relationship objects (product-calcule) 
        if 'PaymentPlan' in class_name_list or 'ContributionPlan' in class_name_list:
            if 'Product' in result_linked_class:
                result_linked_class.remove('Product')
        return LinkedClassListGQLType(result_linked_class)


def _append_to_calcrule_list(list_cr, cr):
    list_cr.append(
        CalculationRulesGQLType(
            calculation_class_name=cr.calculation_rule_name,
            status=cr.status,
            description=cr.description,
            uuid=cr.uuid,
            class_param=cr.impacted_class_parameter,
            date_valid_from=cr.date_valid_from,
            date_valid_to=cr.date_valid_to,
            from_to=cr.from_to,
            type=cr.type,
            sub_type=cr.sub_type
        )
    )
    return list_cr



CLASS_RULE_PARAM_VALIDATION = [
    {
        "class": "ContributionPlan",
        "parameters": [
            {
                "type": "select",
                "name": "rate",
                "label": {
                    "en": "Percentage of income",
                    "fr": "Pourcentage du salaire"
                },
                "rights": {
                    "read": "151201",
                    "write": "151202",
                    "update": "151203",
                    "replace": "151206",
                },
                'relevance': "True",
                'condition': "INPUT>1",
                'optionSet': [
                    {
                        "value": "5",
                        "label": {
                            "en": "5%",
                            "fr": "5%"
                        }
                    },
                    {
                        "value": "10",
                        "label": {
                            "en": "10%",
                            "fr": "10%"
                        }
                    },
                    {
                        "value": "15",
                        "label": {
                            "en": "15%",
                            "fr": "15%"
                        }
                    },
                ],
                "default": "5"
            },
            {
                "type": "checkbox",
                "name": "includeFamilly",
                "label": {
                    "en": "include family members",
                    "fr": "Inclure les membres de la familles"
                },
                "rights": {
                    "read": "151201",
                    "write": "151202",
                    "update": "151203",
                    "replace": "151206",
                },
                "relevance": "True",
                "default": "False"
            },
        ],
    },
    {
        "class": "ContractDetails",
        "parameters": [
            {
                "type": "number",
                "name": "income",
                "label": {
                    "en": "Income",
                    "fr": "Salaire"
                },
                "rights": {
                    "read": "152101",
                    "write": "152102",
                    "update": "152103",
                    "replace": "152103",
                },
                "relevance": "True",
                "condition": "INPUT>100",
                "default": ""
            }
        ],
    },
    {
        "class": "PolicyHolderInsuree",
        "parameters": [
            {
                "type": "number",
                "name": "income",
                "label": {
                    "en": "Income",
                    "fr": "Salaire"
                },
                "rights": {
                    "read": "150201",
                    "write": "150202",
                    "update": "150203",
                    "replace": "150206",
                },
                "relevance": "True",
                "condition": "INPUT>100",
                "default": ""
            }
        ],
    },
]

CLASS_RULE_PARAM_VALIDATION2 = [
    {
        "class": "ContributionPlan",
        "parameters": [
            {
                "type": "select",
                "name": "rate",
                "label": {
                    "en": "Percentage of income",
                    "fr": "Pourcentage du salaire"
                },
                "rights": {
                    "read": "151201",
                    "write": "151202",
                    "update": "151203",
                    "replace": "151206",
                },
                'optionSet': [
                    {
                        "value": "5",
                        "label": {
                            "en": "5%",
                            "fr": "5%"
                        }
                    },
                    {
                        "value": "10",
                        "label": {
                            "en": "10%",
                            "fr": "10%"
                        }
                    },
                    {
                        "value": "15",
                        "label": {
                            "en": "15%",
                            "fr": "15%"
                        }
                    },
                ],
                "default": "5"
            },
        ],
    },
    {
        "class": "ContractDetails",
        "parameters": [
            {
                "type": "number",
                "name": "income",
                "label": {
                    "en": "Income",
                    "fr": "Salaire"
                },
                "rights": {
                    "read": "152101",
                    "write": "152102",
                    "update": "152103",
                    "replace": "152103",
                },
                "default": ""
            }
        ],
    },
    {
        "class": "PolicyHolderInsuree",
        "parameters": [
            {
                "type": "number",
                "name": "income",
                "label": {
                    "en": "Income",
                    "fr": "Salaire"
                },
                "rights": {
                    "read": "150201",
                    "write": "150202",
                    "update": "150203",
                    "replace": "150206",
                },
                "default": ""
            }
        ],
    },
]

DESCRIPTION_CONTRIBUTION_VALUATION = F"" \
    F"This calcutation will add the income in the contract details " \
    F"and PHinsuree and the percentage in the Contribution plan" \
    F" so when a contract valuation is requested then the calculation will" \
    F" determine the value based on the contract details income and CP percentage"

FROM_TO = []



import importlib
import inspect

from core.abs_calculation_rule import AbsCalculationRule
from django.apps import AppConfig


MODULE_NAME = "calculation"


DEFAULT_CFG = {
    "gql_query_calculation_rule_perms": ["153001"],
    "gql_mutation_update_calculation_rule_perms": ["153003"],
}


CALCULATION_RULES = []


def read_all_calculation_rules():
    """function to read all calculation rules"""
    for name, cls in inspect.getmembers(importlib.import_module("calculation.calculation_rule"), inspect.isclass):
        if 'calculation' in cls.__module__.split('.')[0]:
            CALCULATION_RULES.append(cls)
            cls.ready()


class CalculationConfig(AppConfig):
    name = MODULE_NAME

    gql_query_calculation_rule_perms = []
    gql_mutation_update_calculation_rule_perms = []

    def _configure_permissions(self, cfg):
        CalculationConfig.gql_query_calculation_rule_perms = cfg[
            "gql_query_calculation_rule_perms"]
        CalculationConfig.gql_mutation_update_calculation_rule_perms = cfg[
            "gql_mutation_update_calculation_rule_perms"
        ]

    def ready(self):
        from core.models import ModuleConfiguration
        cfg = ModuleConfiguration.get_or_default(MODULE_NAME, DEFAULT_CFG)
        self._configure_permissions(cfg)
        read_all_calculation_rules()
